// File: Algorithm-7-3.cpp
// Stan Warford
// September 2020

#include <cstdlib>
#include <iostream>
#include <thread>
#include <mutex>
#include "Util450.cpp"

using namespace std;

class PCMonitor {
private:
    static const int n = 5;
    int out = 0, in = 0;
    volatile int count = 0;
    int buffer[n];
    mutex pcMutex;
    condition_variable notEmpty;
    condition_variable notFull;

public:
    void append(int v) {
        unique_lock<mutex> guard(pcMutex);
        notFull.wait(guard, [this] { return count != n; });
        buffer[in] = v;
        in = (in + 1) % n;
        count = count + 1;
        cout << "Producer put " << v << endl;
        notEmpty.notify_one();
    }

    int take() {
        unique_lock<mutex> guard(pcMutex);
        notEmpty.wait(guard, [this] { return count != 0; });
        int temp = buffer[out];
        out = (out + 1) % n;
        count = count - 1;
        cout << "Consumer got " << temp << endl;
        notFull.notify_one();
        return temp;
    }
};

PCMonitor pc;

void producerRun() {
    int d;
    cout << "Producer started." << endl;
    for (int i = 0; i < 15; i++) {
        randomDelay(60);
        d = 10 * i;
        pc.append(d);
    }
    cout << "Producer finished." << endl;
}

void consumerRun() {
    int d;
    cout << "Consumer started." << endl;
    for (int i = 0; i < 15; i++) {
        randomDelay(100);
        d = pc.take(); // Ignore returned value
    }
    cout << "Consumer finished." << endl;
}

int main() {
    thread consumer(consumerRun);
    thread producer(producerRun);
    consumer.join();
    producer.join();
    return EXIT_SUCCESS;
}
