// File: ReadersWritersD.cpp
// Readers and writers Algorithm 7.4 with a Mesa semantics monitor
// Deadlocks!
// Stan Warford
// December 2020

#include <cstdlib>
#include <iostream>
#include <thread>
#include <mutex>
#include "Util450.cpp"

using namespace std;

class RWMonitor {
private:
    mutex rwMutex;
    condition_variable okToRead;
    condition_variable okToWrite;
    int readers = 0;
    int writers = 0;
    int blockedOnOKtoRead = 0;
    int blockedOnOKtoWrite = 0;

public:
    void startRead() {
        unique_lock<mutex> guard(rwMutex);
        if (writers == 0 || blockedOnOKtoWrite != 0) {
            blockedOnOKtoRead++;
            okToRead.wait(guard, [this] { return writers != 0 && blockedOnOKtoWrite == 0; });
            blockedOnOKtoRead--;
        }
        readers++;
        okToRead.notify_one();
    }

    void endRead() {
        unique_lock<mutex> guard(rwMutex);
        readers--;
        if (readers == 0) {
            okToWrite.notify_one();
        }
    }

    void startWrite() {
        unique_lock<mutex> guard(rwMutex);
        if (blockedOnOKtoRead == 0) {
            blockedOnOKtoWrite++;
            okToWrite.wait(guard, [this] { return writers == 0 && readers == 0; });
            blockedOnOKtoWrite--;
        }
        writers++;
    }

    void endWrite() {
        unique_lock<mutex> guard(rwMutex);
        if (blockedOnOKtoRead == 0) {
            okToWrite.notify_one();
        } else {
            okToRead.notify_one();
        }
    }
};

class RWDataBase {
private:
    RWMonitor rwMonitor;
    int myData = 5;
    mutex coutMutex;

public:
    void readMyData(int readerID) {
        rwMonitor.startRead();
        coutMutex.lock();
        cout << "Reader " << readerID << " is about to read" << endl;
        coutMutex.unlock();
        randomDelay(60);
        coutMutex.lock();
        cout << "Reader " << readerID << " read " << myData << endl;
        coutMutex.unlock();
        rwMonitor.endRead();
    }

    void writeMyData(int writerID) {
        rwMonitor.startWrite();
        cout << "Writer " << writerID << " is about to write" << endl;
        randomDelay(60);
        myData += 5;
        cout << "Writer " << writerID << " wrote " << myData << endl;
        rwMonitor.endWrite();
    }
};

RWDataBase rwDataBase;

void readerRun(int readerID) {
    for (int i = 0; i < 3; i++) {
        randomDelay(60);
        rwDataBase.readMyData(readerID);
    }
}

void writerRun(int writerID) {
    for (int i = 0; i < 3; i++) {
        randomDelay(60);
        rwDataBase.writeMyData(writerID);
    }
}

int main() {
    thread reader0(readerRun, 0);
    thread reader1(readerRun, 1);
    thread reader2(readerRun, 2);
    thread writer0(writerRun, 0);
    thread writer1(writerRun, 1);
    reader0.join();
    reader1.join();
    reader2.join();
    writer0.join();
    writer1.join();
    return EXIT_SUCCESS;
}
