//
// Created by Stan Warford on 8/30/20.
//
#include <thread>
#include <chrono>
#include <random>
#include <mutex>
#include <iostream>

using namespace std;

random_device rdev{}; // For random seed
default_random_engine engine{rdev()}; // Seed the engine

void randomDelay(int delay) {
    uniform_int_distribution<int> distr(0, delay);
    int d = distr(engine);
    // cout << "delay == " << d << endl;
    this_thread::sleep_for(chrono::milliseconds(d));
}

// Ben-Ari, Principles of Concurrent and Distributed Programming
// Algorithm 7.2: Semaphore simulated with a monitor
class Semaphore {
private:
    int s;
    condition_variable notZero;
    mutex semMutex;

public:

    Semaphore(int k) { s = k; }

    void wait() {
        unique_lock<mutex> guard(semMutex);
        notZero.wait(guard, [this]{return s != 0;});
        s--;
    }

    void signal() {
        lock_guard<mutex> guard(semMutex);
        s++;
        notZero.notify_one();
    }
};
